import "dart:math" as math;

class _VecBase {
  final int x;
  final int y;

  const _VecBase(this.x, this.y);

  _VecBase.fromVector(Vector v)
      : x = v.x,
        y = v.y;

  int get area => x * y;

  int get lengthSquared => x * x + y * y;

  num get length => math.sqrt(this.lengthSquared);

  int get kingLength => math.max(x.abs(), y.abs());

  int get rookLength => x.abs() + y.abs();

  bool operator >(Object o) {
    if (o is _VecBase)
      return lengthSquared > o.lengthSquared;
    else if (o is num) return lengthSquared > o * o;

    throw ArgumentError("Operand's type must be int or Vector");
  }

  bool operator >=(Object o) {
    if (o is _VecBase)
      return lengthSquared >= o.lengthSquared;
    else if (o is num) return lengthSquared >= o * o;

    throw ArgumentError("Operand's type must be int or Vector");
  }

  bool operator <(Object o) {
    if (o is _VecBase)
      return lengthSquared < o.lengthSquared;
    else if (o is num) return lengthSquared < o * o;

    throw ArgumentError("Operand's type must be int or Vector");
  }

  bool operator <=(Object o) {
    if (o is _VecBase)
      return lengthSquared <= o.lengthSquared;
    else if (o is num) return lengthSquared <= o * o;

    throw ArgumentError("Operand's type must be int or Vector");
  }

  Vector operator +(Object o) {
    if (o is _VecBase)
      return Vector(x + o.x, y + o.y);
    else if (o is int) return Vector(x + o, y + o);

    throw ArgumentError("Operand's type must be int or Vector");
  }

  Vector operator -(Object o) {
    if (o is _VecBase)
      return Vector(x - o.x, y - o.y);
    else if (o is int) return Vector(x - o, y - o);

    throw ArgumentError("Operand's type must be int or Vector");
  }

  Vector operator *(int scale) => Vector(x * scale, y * scale);

  Vector operator ~/(int scale) => Vector(x ~/ scale, y ~/ scale);

  Vector abs() => Vector(x.abs(), y.abs());

  Vector offset(int x, int y) => Vector(this.x + x, this.y + y);
  Vector offsetX(int offset) => Vector(x + offset, y);
  Vector offsetY(int offset) => Vector(x, y + offset);

  bool contains(Vector o) {
    var right = math.max(0, x);
    if (o.x > right) return false;

    var left = math.min(0, x);
    if (o.x < left) return false;

    var top = math.max(0, y);
    if (o.y > top) return false;

    var bottom = math.min(0, y);
    if (o.y < bottom) return false;

    return true;
  }

  Direction get nearestDirection {
    if (x == 0) {
      if (y < 0) return Direction.south;
      else if (y == 0) return Direction.none;
      else return Direction.north;
    }

    double slope = y / x;

    if (x < 0) {
      if (slope <= -2.0) return Direction.north;
      else if (slope <= -0.5) return Direction.northWest;
      else if (slope <= 0.5) Direction.west;
      else if (slope <= 2.0) return Direction.southWest;
      else return Direction.south;
    } else {
      if (slope >= 2.0) return Direction.north;
      else if (slope >= 0.5) return Direction.northEast;
      else if (slope >= -0.5) return Direction.east;
      else if (slope >= -2.0) return Direction.southEast;
      else return Direction.south;
    }
  }

  String toString() => "($x, $y)";
}

class Vector extends _VecBase {
  const Vector(int x, int y) : super(x, y);

  bool operator ==(Object o) {
    if (o is _VecBase) return x == o.x && y == o.y;

    return false;
  }
}

class Direction extends _VecBase implements Vector {
  static const Direction none = Direction(0, 0);
  static const Direction north = Direction(0, 1);
  static const Direction northEast = Direction(1, 1);
  static const Direction east = Direction(1, 0);
  static const Direction southEast = Direction(1, -1);
  static const Direction south = Direction(0, -1);
  static const Direction southWest = Direction(-1, -1);
  static const Direction west = Direction(-1, 0);
  static const Direction northWest = Direction(-1, 1);

  static const List<Direction> cardinals = [north, east, south, west];
  static const List<Direction> intercardinals = [
    northEast,
    southEast,
    southWest,
    northWest
  ];
  static const List<Direction> all = [
    north,
    northEast,
    east,
    southEast,
    south,
    southWest,
    west,
    northWest
  ];

  const Direction(int x, int y) : super(x, y);

  ///Rotate [this] direction by `45 degrees` [numOctants] times.
  ///The sign of [numOctants] is direction of rotation:
  ///* positive sign: rotate right.
  ///* negative sign: rotate left.
  Direction rotateByOctants(int numOctants) {
    if (this == none) return none;

    return all[(all.indexOf(this) + numOctants) % all.length];
  }

  Direction rotate45Right() => rotateByOctants(1);
  Direction rotate45Left() => rotateByOctants(-1);
  Direction rotate90Right() => rotateByOctants(2);
  Direction rotate90Left() => rotateByOctants(-2);
  Direction rotate180Right() => rotateByOctants(4);
  Direction rotate180Left() => rotateByOctants(-4);
}
